==> media-0.0.7-sources.jar: commonMain/io/dyte/media/hive/handlers/HiveUnifiedPlan.kt <==
package io.dyte.media.hive.handlers

import io.dyte.media.hive.HiveConnectionState
import io.dyte.webrtc.*
import io.dyte.media.hive.HiveEmitData
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.MutableSharedFlow
import kotlinx.coroutines.flow.launchIn
import kotlinx.coroutines.flow.onEach

class HiveUnifiedPlan(val coroutineScope: CoroutineScope): HiveHandlerInterface() {
  /** Handler Direction */
  private lateinit var _direction: RtpTransceiverDirection

  /** RTCPeerConnection instance */
  private lateinit var _pc: PeerConnection

  /** Local stream for sending general media tracks */
  private lateinit var _sendStream: MediaStream

  /** Local stream for sending screen share tracks */
  private var _sendScreenShareStream: MediaStream? = null

  /** Got transport local and remote parameters */
  private var _transportReady = false

  val observer = MutableSharedFlow<HiveEmitData>()

  init {
    runBlocking {
      _sendStream = MediaDevices.getUserMedia(audio = true, video = true)
      //_sendScreenShareStream = MediaDevices.getDisplayMedia()
    }
  }

  companion object {
    fun createFactory(): HiveHandlerFactory = { HiveUnifiedPlan(coroutineScope = CoroutineScope(Dispatchers.Main)) }
  }

  override fun getName() = "UnifiedPlan"

  override fun getPc() = this._pc

  override fun close() {
    println("DyteMediaClient: UnifiedPlan: close()")

    if (this::_pc.isInitialized) {
      try {
        this._pc.close()
      } catch (e: Error) {
        println("DyteMediaClient: UnifiedPlan: pc.close() $e")
      }
    }
  }

  override suspend fun init(options: HiveHandlerRunOptions) {
    println("DyteMediaClient: UnifiedPlan: init()")

    this._direction = options.direction

    @Suppress("UNCHECKED_CAST")
    this._pc = PeerConnection(RtcConfiguration(
      iceServers = options.additionalSettings?.get("iceServers") as? List<IceServer>
        ?: options.iceServers ?: emptyList(),
      iceTransportPolicy = options.additionalSettings?.get("iceTransportPolicy") as? IceTransportPolicy
        ?: options.iceTransportPolicy ?: IceTransportPolicy.All,
      bundlePolicy = options.additionalSettings?.get("bundlePolicy") as? BundlePolicy ?: BundlePolicy.MaxBundle,
      rtcpMuxPolicy = options.additionalSettings?.get("rtcpMuxPolicy") as? RtcpMuxPolicy  ?: RtcpMuxPolicy.Require,
//      sdpSemantics = "unified-plan"
//      proprietaryConstraints
    ))
    println("HiveDebug: PeerConnection created")


    _pc.onTrack.onEach {
      println("HiveDebug: Received track event")
      options.onTrackHandler(it)
    }.launchIn(coroutineScope)

    _addEventListeners()
  }

  override suspend fun connect(): HiveGenericHandlerResult {
    /** Since we have enabled bundlePolicy to MaxBundle,
     *  we need to have max-bundle attribute in SDP and to generate
     *  that we are using a hacky fix by creating a temp data channel
     */
    println("HiveDebug: Inside handler.connect()")

    val dc = this._pc.createDataChannel("dyte")!!

    println("HiveDebug: Data channel created $dc")

    val offer = this._pc.createOffer(OfferAnswerOptions())

    println("DyteMediaClient: UnifiedPlan: connect  offer with sdp = ${offer.sdp}")

    this._pc.setLocalDescription(offer)

    val callback: suspend (SessionDescription) -> Unit = { answer ->
      println("DyteMediaClient: UnifiedPlan: connect(): Calling pc.setRemoteDescription() with answer=${answer.sdp}")
      println("HiveDebug: pc = $_pc dc = $dc")

      this._pc.setRemoteDescription(answer)
      println("HiveDebug: remote description set")
      dc.close()
      println("HiveDebug: data channel closed")
    }

    return HiveGenericHandlerResult(
      offerSdp = offer,
      callback = callback
    )
  }

  override suspend fun updateIceServers(iceServers: List<IceServer>) {
    println("DyteMediaClient: UnifiedPlan: updateIceServers()")

//    val configuration = this._pc.getConfiguration()
//    configuration.iceServers = iceServers
//    this._pc.setConfiguration(configuration)

    // TODO: Modify to retain other configuration parameters or implement getConfiguration
    this._pc.setConfiguration(RtcConfiguration(iceServers = iceServers))
  }

  override suspend fun restartIce(): HiveGenericHandlerResult {
    println("DyteMediaClient: UnifiedPlan: restartIce()")

    val offer = this._pc.createOffer(OfferAnswerOptions(iceRestart = true))

    val callback: suspend (SessionDescription) -> Unit = { answer ->
      println("DyteMediaClient: UnifiedPlan: restartIce(): Calling pc.setRemoteDescription() with answer=${answer.sdp}")

      this._pc.setRemoteDescription(answer)
    }

    return HiveGenericHandlerResult(
      offerSdp = offer,
      callback = callback
    )
  }

  override suspend fun getTransportStats(): RtcStatsReport = this._pc.getStats()!!

  override suspend fun send(options: HiveHandlerSendOptions): HiveHandlerSendResult {
    println("ProducerDebug: Inside handler.send()")
    this._assertSendDirection()
    println("ProducerDebug: asserted send direction")

    println("ProducerDebug: UnifiedPlan: send() with kind = ${options.track.kind} with trackId = ${options.track.id}")

    println("ProducerDebug: UnifiedPlan: send(): Creating new transceiver")

    // TODO: Return the created transceiver from addTransceiver() itself
    this._pc.addTransceiver(
      track = options.track,
      kind = options.track.kind,
      init = CommonRtpTransceiverInit(
        direction = RtpTransceiverDirection.SendOnly,
        streams = listOf(this._sendStream),
        sendEncodings = options.encodings ?: emptyList()
      )
    )

    println("ProducerDebug: Transceiver added")

    val trans = this._pc.getTransceivers().last()

    println("ProducerDebug: Found transceiver")

    // Note(anunaym14): Similar to Firefox, WebRTC-KMP doesn't support RtpTransceiver.setCodecPreferences()
    // So, let's enforce SFU to only use VP8 in server side.

    val offer = this._pc.createOffer(OfferAnswerOptions())

    println("ProducerDebug: Offer sdp for ${options.track.kind} = ${offer.sdp}")

    //if (!this._transportReady) throw Error("WebRTC Transport not connected")

    println("ProducerDebug: Transport ready verified")

    this._pc.setLocalDescription(offer)

    println("ProducerDebug: Local description set")

    var updatedOffer: SessionDescription? = null

    /** mangle offer with new fmtp parameters and nack*/
    if (options.codecOptions != null && options.codecOptions.name == "opus") {
      var updatedOfferSDP = offer.sdp.replace("minptime=10;useinbandfec=1",
        "minptime=10;useinbandfec=1;usedtx=1")

      updatedOfferSDP += "a=rtcp-fb:111 nack\r\n"

       updatedOffer = SessionDescription(
        type = offer.type,
        sdp = updatedOfferSDP
      )
    }

    val setAnswer: suspend (SessionDescription) -> String = { answer ->
      println("ProducerDebug: UnifiedPlan: send(): Calling pc.setRemoteDescription() " +
              "for ${options.track.kind} with answer=${answer.sdp}")

      this._pc.setRemoteDescription(answer)

      // Store in the map
      this.mapMidTransceiver[trans.mid] = trans

      trans.mid
    }

    return HiveHandlerSendResult(
      offerSdp = updatedOffer ?: offer,
      callback = setAnswer
    )
  }

  override suspend fun stopSending(localId: String): HiveGenericHandlerResult {
    this._assertSendDirection()

    println("DyteMediaClient: UnifiedPlan: stopSending for localId: $localId")

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    transceiver.sender.replaceTrack(null)
    this._pc.removeTrack(transceiver.sender)

    transceiver.direction = RtpTransceiverDirection.Inactive

    val offer = this._pc.createOffer(OfferAnswerOptions())

    println("DyteMediaClient: calling pc.setLocalDescription() with offer = ${offer.sdp}")

    this._pc.setLocalDescription(offer)

    val callback: suspend (SessionDescription) -> Unit = { answer ->
      println("DyteMediaClient: UnifiedPlan: stopSending(): Calling pc.setRemoteDescription() with answer=${answer.sdp}")

      this._pc.setRemoteDescription(answer)

      this.mapMidTransceiver.remove(localId)
    }

    return HiveGenericHandlerResult(
      offerSdp = offer,
      callback = callback
    )
  }

  override suspend fun replaceTrack(localId: String, track: MediaStreamTrack?) {
    this._assertSendDirection()

    if (track != null) println("DyteMediaClient: UnifiedPlan: replaceTrack() with localId=$localId, trackId=${track.id}")
    else println("DyteMediaClient: UnifiedPlan: replaceTrack() localId=$localId with no track")

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    transceiver.sender.replaceTrack(track)
  }

  override suspend fun setMaxSpatialLayer(localId: String, spatialLayer: Long) {
    this._assertSendDirection()

    println("DyteMediaClient: UnifiedPlan: setMaxSpatialLayer() with localId=$localId spatialLayer=$spatialLayer")

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    val parameters = transceiver.sender.parameters

    parameters.encodings.forEachIndexed { idx, encoding ->
      // TODO: Verify if changing active to var changes it internally
      encoding.active = idx <= spatialLayer
    }

    transceiver.sender.parameters = parameters
  }

  override suspend fun setRtpEncodingParameters(localId: String, params: Any) {
    this._assertSendDirection()

    println("DyteMediaClient: UnifiedPlan: setRtpEncodingParameters() with localId=$localId params=$params")

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    val parameters = transceiver.sender.parameters

    // TODO: This is required for simulcast which Hive doesn't support right now, fix for future
    // Also, figure out how to use common encodings here
    //parameters.encodings.add(params as RtpEncodingParameters)

    transceiver.sender.parameters = parameters
  }

  override fun getSenderStats(localId: String): List<RtcStatsReport> {
    this._assertSendDirection()

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    // TODO: Implement sender getStats()
//    return transceiver.sender.getStats()
    return emptyList()
  }

  override suspend fun stopReceiving(localId: String): HiveGenericHandlerResult {
    this._assertRecvDirection()

    println("DyteMediaClient: UnifiedPlan: stopReceiving() with localId=$localId")

    throw Error("Method not implemented")
  }

  override suspend fun pauseReceiving(localId: String): HiveGenericHandlerResult {
    this._assertRecvDirection()

    println("DyteMediaClient: UnifiedPlan: pauseReceiving() with localId=$localId")

    throw Error("Method not implemented")
  }

  override suspend fun resumeReceiving(localId: String): HiveGenericHandlerResult {
    this._assertRecvDirection()

    println("DyteMediaClient: UnifiedPlan: resumeReceiving() with localId=$localId")

    throw Error("Method not implemented")
  }

  override suspend fun getReceiverStats(localId: String): List<RtcStatsReport> {
    this._assertRecvDirection()

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    // TODO: Implement receiver getStats()
//    return transceiver.receiver.getStats()
    return emptyList()
  }

  private fun _assertSendDirection() {
    if (this._direction != RtpTransceiverDirection.SendOnly) {
      throw Error("Method can just be called for handlers with send direction")
    }
  }

  private fun _assertRecvDirection() {
    if (this._direction != RtpTransceiverDirection.RecvOnly) {
      throw Error("Method can just be called for handlers with recv direction")
    }
  }

  private suspend fun _generateOffer(): SessionDescription {
    val offer = this._pc.createOffer(OfferAnswerOptions())

    this._pc.setLocalDescription(offer)

    return offer
  }

  private suspend fun _setAnswer(answer: SessionDescription) {
    println("DyteMediaClient: UnifiedPlan: _setAnswer() | calling pc.setRemoteDescription() with answer=${answer.sdp}")

    this._pc.setRemoteDescription(answer)
  }

  private suspend fun _addEventListeners() {
    var job: Job? = null

    this._pc.onIceConnectionStateChange.onEach {
      println("HiveDebug: Handler: IceConnectionStateChange: $it")
      when (it) {
        IceConnectionState.Checking -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Connecting
        ))

        IceConnectionState.Connected -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Connected
        ))

        IceConnectionState.Completed -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Connected
        ))

        IceConnectionState.Failed -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Failed
        ))

        IceConnectionState.Disconnected -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Disconnected
        ))

        IceConnectionState.Closed -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Closed
        ))

        else -> println("Unknown IceConnectionState")
      }
    }.launchIn(coroutineScope)

    this._pc.onIceCandidate.onEach {
      println("HiveDebug: Handler: IceCandidate")
      this.observer.emit(HiveEmitData(
        eventName = "@icecandidate",
        data = it
      ))
    }.launchIn(coroutineScope)

    this._pc.onNegotiationNeeded.onEach {
      println("HiveDebug: Handler: NegotiationNeeded")
      this.observer.emit(HiveEmitData("@negotiationneeded"))
    }.launchIn(coroutineScope)

    this._pc.onIceGatheringState.onEach {
      println("HiveDebug: Handler: IceGatheringState")
      when (it) {
        IceGatheringState.Gathering -> {
          println("IceGatheringStateChange | Gathering")

          this.observer.emit(HiveEmitData(
            eventName = "@icegatheringstatechange",
            data = IceGatheringState.Gathering
          ))
        }

        IceGatheringState.Complete -> {
          println("IceGatheringStateChange | Complete")

          this.observer.emit(HiveEmitData(
            eventName = "@icegatheringstatechange",
            data = IceGatheringState.Complete
          ))
        }

        else -> println("Unknown IceGatheringState")
      }
    }.launchIn(coroutineScope)

    // this._pc.onIceCandidateError.collect {}

    this._pc.onDataChannel.onEach {channel ->
      println("DCDebug: HiveDebug: Data channel created: $channel ${channel.id} ${channel.label} ${channel.readyState} for ${this._direction}")

      channel.onOpen.onEach {
        println("DCDebug: HiveDebug: Data channel open: ${channel.label}")
      }.launchIn(coroutineScope)

      channel.onClose.onEach {
        println("DCDebug: HiveDebug: Data channel closed: ${channel.label}")
      }.launchIn(coroutineScope)

      channel.onError.onEach {error ->
        println("DCDebug: HiveDebug: Data channel error: $error for ${channel.label}")
      }.launchIn(coroutineScope)

      channel.onMessage.onEach {msg ->
        println("DCDebug: HiveDebug: Data channel message: $msg for ${channel.label}")

        this.observer.emit(HiveEmitData(
          eventName = "datachannel",
          data = mapOf(
            "channel" to channel,
            "message" to msg.decodeToString()
          )
        ))
      }.launchIn(coroutineScope)
    }.launchIn(coroutineScope)
  }
}

==> media-0.0.10-sources.jar: commonMain/io/dyte/media/hive/handlers/HiveUnifiedPlan.kt <==
package io.dyte.media.hive.handlers

import io.dyte.media.hive.HiveConnectionState
import io.dyte.webrtc.*
import io.dyte.media.hive.HiveEmitData
import io.dyte.media.utils.IMediaClientLogger
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.MutableSharedFlow
import kotlinx.coroutines.flow.launchIn
import kotlinx.coroutines.flow.onEach

class HiveUnifiedPlan(
  val coroutineScope: CoroutineScope,
  private val logger: IMediaClientLogger?
): HiveHandlerInterface() {
  /** Handler Direction */
  private lateinit var _direction: RtpTransceiverDirection

  /** RTCPeerConnection instance */
  private lateinit var _pc: PeerConnection

  /** Got transport local and remote parameters */
  private var _transportReady = false

  val observer = MutableSharedFlow<HiveEmitData>()

  companion object {
    fun createFactory(logger: IMediaClientLogger? = null): HiveHandlerFactory = { 
      HiveUnifiedPlan(
        coroutineScope = CoroutineScope(Dispatchers.Main),
        logger = logger
      ) 
    }
  }

  override fun getName() = "UnifiedPlan"

  override fun getPc() = this._pc

  override fun close() {
    logger?.traceLog("DyteMediaClient: UnifiedPlan: close()")

    if (this::_pc.isInitialized) {
      try {
        this._pc.close()
      } catch (e: Error) {
        logger?.traceLog("DyteMediaClient: UnifiedPlan: pc.close() $e")
      }
    }
  }

  override suspend fun init(options: HiveHandlerRunOptions) {
    logger?.traceLog("DyteMediaClient: UnifiedPlan: init()")

    this._direction = options.direction

    @Suppress("UNCHECKED_CAST")
    this._pc = PeerConnection(RtcConfiguration(
      iceServers = options.additionalSettings?.get("iceServers") as? List<IceServer>
        ?: options.iceServers ?: emptyList(),
      iceTransportPolicy = options.additionalSettings?.get("iceTransportPolicy") as? IceTransportPolicy
        ?: options.iceTransportPolicy ?: IceTransportPolicy.All,
      bundlePolicy = options.additionalSettings?.get("bundlePolicy") as? BundlePolicy ?: BundlePolicy.MaxBundle,
      rtcpMuxPolicy = options.additionalSettings?.get("rtcpMuxPolicy") as? RtcpMuxPolicy  ?: RtcpMuxPolicy.Require,
//      sdpSemantics = "unified-plan"
//      proprietaryConstraints
    ))
    logger?.traceLog("HiveDebug: PeerConnection created")


    _pc.onTrack.onEach {
      logger?.traceLog("HiveDebug: Received track event")
      options.onTrackHandler(it)
    }.launchIn(coroutineScope)

    _addEventListeners()
  }

  override suspend fun connect(): HiveGenericHandlerResult {
    /** Since we have enabled bundlePolicy to MaxBundle,
     *  we need to have max-bundle attribute in SDP and to generate
     *  that we are using a hacky fix by creating a temp data channel
     */
    logger?.traceLog("HiveDebug: Inside handler.connect()")

    val dc = this._pc.createDataChannel("dyte")!!

    logger?.traceLog("HiveDebug: Data channel created $dc")

    val offer = this._pc.createOffer(OfferAnswerOptions())

    logger?.traceLog("DyteMediaClient: UnifiedPlan: connect  offer with sdp = ${offer.sdp}")

    this._pc.setLocalDescription(offer)

    val callback: suspend (SessionDescription) -> Unit = { answer ->
      logger?.traceLog("DyteMediaClient: UnifiedPlan: connect(): Calling pc.setRemoteDescription() with answer=${answer.sdp}")
      logger?.traceLog("HiveDebug: pc = $_pc dc = $dc")

      this._pc.setRemoteDescription(answer)
      logger?.traceLog("HiveDebug: remote description set")
      dc.close()
      logger?.traceLog("HiveDebug: data channel closed")
    }

    return HiveGenericHandlerResult(
      offerSdp = offer,
      callback = callback
    )
  }

  override suspend fun updateIceServers(iceServers: List<IceServer>) {
    logger?.traceLog("DyteMediaClient: UnifiedPlan: updateIceServers()")

//    val configuration = this._pc.getConfiguration()
//    configuration.iceServers = iceServers
//    this._pc.setConfiguration(configuration)

    // TODO: Modify to retain other configuration parameters or implement getConfiguration
    this._pc.setConfiguration(RtcConfiguration(iceServers = iceServers))
  }

  override suspend fun restartIce(): HiveGenericHandlerResult {
    logger?.traceLog("DyteMediaClient: UnifiedPlan: restartIce()")

    val offer = this._pc.createOffer(OfferAnswerOptions(iceRestart = true))

    val callback: suspend (SessionDescription) -> Unit = { answer ->
      logger?.traceLog("DyteMediaClient: UnifiedPlan: restartIce(): Calling pc.setRemoteDescription() with answer=${answer.sdp}")

      this._pc.setRemoteDescription(answer)
    }

    return HiveGenericHandlerResult(
      offerSdp = offer,
      callback = callback
    )
  }

  override suspend fun getTransportStats(): RtcStatsReport = this._pc.getStats()!!

  override suspend fun send(options: HiveHandlerSendOptions): HiveHandlerSendResult {
    logger?.traceLog("ProducerDebug: Inside handler.send()")
    this._assertSendDirection()
    logger?.traceLog("ProducerDebug: asserted send direction")

    logger?.traceLog("ProducerDebug: UnifiedPlan: send() with kind = ${options.track.kind} with trackId = ${options.track.id}")

    logger?.traceLog("ProducerDebug: UnifiedPlan: send(): Creating new transceiver")

    // TODO: Return the created transceiver from addTransceiver() itself
    this._pc.addTransceiver(
      track = options.track,
      kind = options.track.kind,
      init = CommonRtpTransceiverInit(
        direction = RtpTransceiverDirection.SendOnly,
        streams = listOf(options.stream),
        sendEncodings = options.encodings ?: emptyList()
      )
    )

    logger?.traceLog("ProducerDebug: Transceiver added")

    val trans = this._pc.getTransceivers().last()

    logger?.traceLog("ProducerDebug: Found transceiver")

    // Note(anunaym14): Similar to Firefox, WebRTC-KMP doesn't support RtpTransceiver.setCodecPreferences()
    // So, let's enforce SFU to only use VP8 in server side.

    val offer = this._pc.createOffer(OfferAnswerOptions())

    logger?.traceLog("ProducerDebug: Offer sdp for ${options.track.kind} = ${offer.sdp}")

    //if (!this._transportReady) throw Error("WebRTC Transport not connected")

    logger?.traceLog("ProducerDebug: Transport ready verified")

    this._pc.setLocalDescription(offer)

    logger?.traceLog("ProducerDebug: Local description set")

    var updatedOffer: SessionDescription? = null

    /** mangle offer with new fmtp parameters and nack*/
    if (options.codecOptions != null && options.codecOptions.name == "opus") {
      var updatedOfferSDP = offer.sdp.replace("minptime=10;useinbandfec=1",
        "minptime=10;useinbandfec=1;usedtx=1")

      updatedOfferSDP += "a=rtcp-fb:111 nack\r\n"

       updatedOffer = SessionDescription(
        type = offer.type,
        sdp = updatedOfferSDP
      )
    }

    val setAnswer: suspend (SessionDescription) -> String = { answer ->
      logger?.traceLog("ProducerDebug: UnifiedPlan: send(): Calling pc.setRemoteDescription() " +
              "for ${options.track.kind} with answer=${answer.sdp}")

      this._pc.setRemoteDescription(answer)

      // Store in the map
      this.mapMidTransceiver[trans.mid] = trans

      trans.mid
    }

    return HiveHandlerSendResult(
      offerSdp = updatedOffer ?: offer,
      callback = setAnswer
    )
  }

  override suspend fun stopSending(localId: String): HiveGenericHandlerResult {
    this._assertSendDirection()

    logger?.traceLog("DyteMediaClient: UnifiedPlan: stopSending for localId: $localId")

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    transceiver.sender.replaceTrack(null)
    this._pc.removeTrack(transceiver.sender)

    transceiver.direction = RtpTransceiverDirection.Inactive

    val offer = this._pc.createOffer(OfferAnswerOptions())

    logger?.traceLog("DyteMediaClient: calling pc.setLocalDescription() with offer = ${offer.sdp}")

    this._pc.setLocalDescription(offer)

    val callback: suspend (SessionDescription) -> Unit = { answer ->
      logger?.traceLog("DyteMediaClient: UnifiedPlan: stopSending(): Calling pc.setRemoteDescription() with answer=${answer.sdp}")

      this._pc.setRemoteDescription(answer)

      this.mapMidTransceiver.remove(localId)
    }

    return HiveGenericHandlerResult(
      offerSdp = offer,
      callback = callback
    )
  }

  override suspend fun replaceTrack(localId: String, track: MediaStreamTrack?) {
    this._assertSendDirection()

    if (track != null) logger?.traceLog("DyteMediaClient: UnifiedPlan: replaceTrack() with localId=$localId, trackId=${track.id}")
    else logger?.traceLog("DyteMediaClient: UnifiedPlan: replaceTrack() localId=$localId with no track")

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    transceiver.sender.replaceTrack(track)
  }

  override suspend fun setMaxSpatialLayer(localId: String, spatialLayer: Long) {
    this._assertSendDirection()

    logger?.traceLog("DyteMediaClient: UnifiedPlan: setMaxSpatialLayer() with localId=$localId spatialLayer=$spatialLayer")

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    val parameters = transceiver.sender.parameters

    parameters.encodings.forEachIndexed { idx, encoding ->
      // TODO: Verify if changing active to var changes it internally
      encoding.active = idx <= spatialLayer
    }

    transceiver.sender.parameters = parameters
  }

  override suspend fun setRtpEncodingParameters(localId: String, params: Any) {
    this._assertSendDirection()

    logger?.traceLog("DyteMediaClient: UnifiedPlan: setRtpEncodingParameters() with localId=$localId params=$params")

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    val parameters = transceiver.sender.parameters

    // TODO: This is required for simulcast which Hive doesn't support right now, fix for future
    // Also, figure out how to use common encodings here
    //parameters.encodings.add(params as RtpEncodingParameters)

    transceiver.sender.parameters = parameters
  }

  override fun getSenderStats(localId: String): List<RtcStatsReport> {
    this._assertSendDirection()

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    // TODO: Implement sender getStats()
//    return transceiver.sender.getStats()
    return emptyList()
  }

  override suspend fun stopReceiving(localId: String): HiveGenericHandlerResult {
    this._assertRecvDirection()

    logger?.traceLog("DyteMediaClient: UnifiedPlan: stopReceiving() with localId=$localId")

    throw Error("Method not implemented")
  }

  override suspend fun pauseReceiving(localId: String): HiveGenericHandlerResult {
    this._assertRecvDirection()

    logger?.traceLog("DyteMediaClient: UnifiedPlan: pauseReceiving() with localId=$localId")

    throw Error("Method not implemented")
  }

  override suspend fun resumeReceiving(localId: String): HiveGenericHandlerResult {
    this._assertRecvDirection()

    logger?.traceLog("DyteMediaClient: UnifiedPlan: resumeReceiving() with localId=$localId")

    throw Error("Method not implemented")
  }

  override suspend fun getReceiverStats(localId: String): List<RtcStatsReport> {
    this._assertRecvDirection()

    val transceiver = this.mapMidTransceiver[localId] ?: throw Error("Associated RtpTransceiver not found")

    // TODO: Implement receiver getStats()
//    return transceiver.receiver.getStats()
    return emptyList()
  }

  private fun _assertSendDirection() {
    if (this._direction != RtpTransceiverDirection.SendOnly) {
      throw Error("Method can just be called for handlers with send direction")
    }
  }

  private fun _assertRecvDirection() {
    if (this._direction != RtpTransceiverDirection.RecvOnly) {
      throw Error("Method can just be called for handlers with recv direction")
    }
  }

  private suspend fun _generateOffer(): SessionDescription {
    val offer = this._pc.createOffer(OfferAnswerOptions())

    this._pc.setLocalDescription(offer)

    return offer
  }

  private suspend fun _setAnswer(answer: SessionDescription) {
    logger?.traceLog("DyteMediaClient: UnifiedPlan: _setAnswer() | calling pc.setRemoteDescription() with answer=${answer.sdp}")

    this._pc.setRemoteDescription(answer)
  }

  private suspend fun _addEventListeners() {
    var job: Job? = null

    this._pc.onIceConnectionStateChange.onEach {
      logger?.traceLog("HiveDebug: Handler: IceConnectionStateChange: $it")
      when (it) {
        IceConnectionState.Checking -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Connecting
        ))

        IceConnectionState.Connected -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Connected
        ))

        IceConnectionState.Completed -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Connected
        ))

        IceConnectionState.Failed -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Failed
        ))

        IceConnectionState.Disconnected -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Disconnected
        ))

        IceConnectionState.Closed -> this.observer.emit(HiveEmitData(
          eventName = "@connectionstatechange",
          data = HiveConnectionState.Closed
        ))

        else -> logger?.traceLog("Unknown IceConnectionState")
      }
    }.launchIn(coroutineScope)

    this._pc.onIceCandidate.onEach {
      logger?.traceLog("HiveDebug: Handler: IceCandidate")
      this.observer.emit(HiveEmitData(
        eventName = "@icecandidate",
        data = it
      ))
    }.launchIn(coroutineScope)

    this._pc.onNegotiationNeeded.onEach {
      logger?.traceLog("HiveDebug: Handler: NegotiationNeeded")
      this.observer.emit(HiveEmitData("@negotiationneeded"))
    }.launchIn(coroutineScope)

    this._pc.onIceGatheringState.onEach {
      logger?.traceLog("HiveDebug: Handler: IceGatheringState")
      when (it) {
        IceGatheringState.Gathering -> {
          logger?.traceLog("IceGatheringStateChange | Gathering")

          this.observer.emit(HiveEmitData(
            eventName = "@icegatheringstatechange",
            data = IceGatheringState.Gathering
          ))
        }

        IceGatheringState.Complete -> {
          logger?.traceLog("IceGatheringStateChange | Complete")

          this.observer.emit(HiveEmitData(
            eventName = "@icegatheringstatechange",
            data = IceGatheringState.Complete
          ))
        }

        else -> logger?.traceLog("Unknown IceGatheringState")
      }
    }.launchIn(coroutineScope)

    // this._pc.onIceCandidateError.collect {}

    this._pc.onDataChannel.onEach {channel ->
      logger?.traceLog("DCDebug: HiveDebug: Data channel created: $channel ${channel.id} ${channel.label} ${channel.readyState} for ${this._direction}")

      channel.onOpen.onEach {
        logger?.traceLog("DCDebug: HiveDebug: Data channel open: ${channel.label}")
      }.launchIn(coroutineScope)

      channel.onClose.onEach {
        logger?.traceLog("DCDebug: HiveDebug: Data channel closed: ${channel.label}")
      }.launchIn(coroutineScope)

      channel.onError.onEach {error ->
        logger?.traceLog("DCDebug: HiveDebug: Data channel error: $error for ${channel.label}")
      }.launchIn(coroutineScope)

      channel.onMessage.onEach {msg ->
        logger?.traceLog("DCDebug: HiveDebug: Data channel message: $msg for ${channel.label}")

        this.observer.emit(HiveEmitData(
          eventName = "datachannel",
          data = mapOf(
            "channel" to channel,
            "message" to msg.decodeToString()
          )
        ))
      }.launchIn(coroutineScope)
    }.launchIn(coroutineScope)
  }
}
